"""Unittests for emoji.unicode_codes."""


import emoji


# Build all language packs (i.e. fill the cache):
emoji.emojize("", language="alias")
for lang_code in emoji.LANGUAGES:
    emoji.emojize("", language=lang_code)


def test_emoji_english_names():

    for language, group in (
            ('en', emoji.unicode_codes._EMOJI_UNICODE['en']),
            ('alias', emoji.unicode_codes._ALIASES_UNICODE)
    ):
        for name, ucode in group.items():
            assert name.startswith(':') and name.endswith(':') and len(name) >= 3
            emj = emoji.emojize(name, language=language)
            assert emj == ucode, '%s != %s' % (emoji.emojize(name), ucode)


def test_compare_normal_and_aliases():
    # There should always be more aliases than normal codes
    # since the aliases contain the normal codes

    assert len(emoji.unicode_codes._EMOJI_UNICODE['en']) < len(
        emoji.unicode_codes._ALIASES_UNICODE)


def test_no_alias_duplicates():
    # There should not be two emoji with the same alias
    # (aliases still can be the same as another 'en'-name)
    all_aliases = set()
    for data in emoji.EMOJI_DATA.values():
        if data['status'] <= emoji.STATUS['fully_qualified'] and 'alias' in data:
            for alias in data['alias']:
                assert alias not in all_aliases
                all_aliases.add(alias)



from emoji.unicode_codes.data_dict import *

__all__ = [
    'get_emoji_unicode_dict', 'get_aliases_unicode_dict',
    'EMOJI_DATA', 'STATUS', 'LANGUAGES'
]


_EMOJI_UNICODE = {lang: None for lang in LANGUAGES}  # Cache for the language dicts

_ALIASES_UNICODE = {}  # Cache for the aliases dict


def get_emoji_unicode_dict(lang):
    """Generate dict containing all fully-qualified and component emoji name for a language
    The dict is only generated once per language and then cached in _EMOJI_UNICODE[lang]"""

    if _EMOJI_UNICODE[lang] is None:
        _EMOJI_UNICODE[lang] = {data[lang]: emj for emj, data in EMOJI_DATA.items()
                                if lang in data and data['status'] <= STATUS['fully_qualified']}

    return _EMOJI_UNICODE[lang]


def get_aliases_unicode_dict():
    """Generate dict containing all fully-qualified and component aliases
    The dict is only generated once and then cached in _ALIASES_UNICODE"""

    if not _ALIASES_UNICODE:
        _ALIASES_UNICODE.update(get_emoji_unicode_dict('en'))
        for emj, data in EMOJI_DATA.items():
            if 'alias' in data and data['status'] <= STATUS['fully_qualified']:
                for alias in data['alias']:
                    _ALIASES_UNICODE[alias] = emj

    return _ALIASES_UNICODE



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

import emoji as emoji_pkg
project = 'emoji'
copyright = emoji_pkg.__source__
author = emoji_pkg.__author__

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = emoji_pkg.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

html_baseurl = 'https://carpedm20.github.io/emoji/docs/'

html_theme_options = {
    'page_width': '1200px',
    'github_user': 'carpedm20',
    'github_repo': 'emoji',
    'github_banner': False,
    'show_related': False,
    'canonical_url': 'https://carpedm20.github.io/emoji/docs/',
    'extra_nav_links': {
        'GitHub repository': 'https://github.com/carpedm20/emoji',
        'emoji on PyPI': 'https://pypi.org/project/emoji/',
        'All supported emoji': 'https://carpedm20.github.io/emoji/',
        'All emoji on unicode.org': 'http://www.unicode.org/emoji/charts/full-emoji-list.html',
        'Unicode Standard': 'http://www.unicode.org/reports/tr51/',
    }
}


from .core import (
    demojize as demojize,
    distinct_emoji_list as distinct_emoji_list,
    emoji_count as emoji_count,
    emoji_list as emoji_list,
    emojize as emojize,
    is_emoji as is_emoji,
    replace_emoji as replace_emoji,
    version as version,
    analyze as analyze,
    config as config,
)
from .tokenizer import (
    Token as Token,
    EmojiMatch as EmojiMatch,
    EmojiMatchZWJ as EmojiMatchZWJ,
    EmojiMatchZWJNonRGI as EmojiMatchZWJNonRGI,
)


from .unicode_codes import EMOJI_DATA, LANGUAGES, STATUS

__all__ = [
    # emoji.core
    'emojize', 'demojize', 'analyze', 'config',
    'emoji_list', 'distinct_emoji_list', 'emoji_count',
    'replace_emoji', 'is_emoji', 'version',
    'Token', 'EmojiMatch', 'EmojiMatchZWJ', 'EmojiMatchZWJNonRGI',
    # emoji.unicode_codes
    'EMOJI_DATA', 'STATUS', 'LANGUAGES',
]

__version__: str
__author__: str
__email__: str
__source__: str
__license__: str


from typing import NamedTuple, Union, Dict, Iterator, Any

_SearchTree = Dict[str, Union['_SearchTree', dict[str, dict[str, Any]]]]

_SEARCH_TREE: _SearchTree


class EmojiMatch:
    emoji: str
    start: int
    end: int
    data: dict[str, Any] | None
    def __init__(self, emoji: str, start: int,
                 end: int, data: dict | None): ...

    def data_copy(self) -> Dict[str, Any]: ...
    def is_zwj(self) -> bool: ...
    def split(self) -> EmojiMatchZWJ | EmojiMatch: ...
    def __repr__(self) -> str: ...


class EmojiMatchZWJ(EmojiMatch):
    def __init__(self, match: EmojiMatch): ...
    def join(self) -> str: ...
    def is_zwj(self) -> bool: ...
    def split(self) -> EmojiMatchZWJ: ...
    def __repr__(self) -> str: ...


class EmojiMatchZWJNonRGI(EmojiMatchZWJ):
    def __init__(self, first_emoji_match: EmojiMatch,
                 second_emoji_match: EmojiMatch): ...


class Token(NamedTuple):
    chars: str
    value: str | EmojiMatch


def tokenize(string, keep_zwj: bool) -> Iterator[Token]: ...


def filter_tokens(matches: Iterator[Token], emoji_only: bool,
                  join_emoji: bool) -> Iterator[Token]: ...


def get_search_tree() -> _SearchTree: ...



"""
Extract the full list of emoji and names from the Unicode emoji data files
https://unicode.org/Public/emoji/{v}/emoji-test.txt and
https://www.unicode.org/Public/{v}/ucd/emoji/emoji-variation-sequences.txt
and apply as much formatting as possible so the codes can be dropped into the
emoji registry file.
"""

import sys
import os
import unicodedata
import re
import requests
import bs4
import xml.etree.ElementTree as ET
import logging

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)

include = os.path.relpath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, include)
import emoji as emoji_pkg


def get_text_from_url(url: str) -> str:
    """Get text from url"""

    return requests.get(url).text


def get_emoji_from_url(version: float) -> list:
    """Get splitlines of emojis list from unicode.org"""

    url = f"https://unicode.org/Public/emoji/{version}/emoji-test.txt"
    return get_text_from_url(url).splitlines()


def get_emoji_variation_sequence_from_url(version: str) -> list:
    """Get splitlines of emoji variation sequences from unicode.org"""

    url = f"https://www.unicode.org/Public/{version}/ucd/emoji/emoji-variation-sequences.txt"
    return get_text_from_url(url).splitlines()


def get_emojiterra_from_url(url: str) -> dict:
    html = get_text_from_url(url)

    soup = bs4.BeautifulSoup(html, "html.parser")
    emojis = {}

    data = soup.find_all('li')
    data = [i for i in data if 'href' not in i.attrs and 'data-e' in i.attrs and i['data-e'].strip()]

    for i in data:
        code = i['data-e']
        emojis[code] = i['title'].strip()

    assert len(data) > 100, f"emojiterra data from {url} has only {len(data)} entries"

    return emojis


def get_cheat_sheet(url: str) -> dict:
    """
    Returns a dict of emoji to short-names:
    E.g. {'👴': ':old_man:', '👵': ':old_woman:', ... }
    """

    html = get_text_from_url(url)

    soup = bs4.BeautifulSoup(html, "html.parser")
    emojis = {}

    items = soup.find(class_='ecs-list').find_all(class_='_item')

    pattern = re.compile(r'U\+([0-9A-F]+)')

    for i in items:
        unicode_text = i.find(class_='unicode').text

        code_points = pattern.findall(unicode_text)
        code = ''.join(chr(int(x,16)) for x in code_points)

        emojis[code] = i.find(class_='shortcode').text

    # Remove some unwanted and some weird entries from the cheat sheet
    filtered = {}
    for emj, short_code in emojis.items():

        if short_code.startswith(':flag_'):
            # Skip flags from cheat-sheet, because we already have very similar aliases for the flags
            continue

        if '⊛' in short_code:
            # Strange emoji with ⊛ in the short-code
            continue

        if emj == '\U0001F93E\U0000200D\U00002640\U0000FE0F':
            # The short-code for this emoji is wrong
            continue

        if emj == '\U0001F468\U0000200D\U0001F468\U0000200D\U0001F467':
            # The short-code for this emoji is wrong
            continue

        if short_code.startswith('::'):
            # Do not allow short-codes to have double :: at the start
            short_code = short_code[1:]

        if short_code.endswith('::'):
            # Do not allow short-codes to have double :: at the end
            short_code = short_code[:-1]

        filtered[emj] = short_code

    assert len(filtered) > 100, f"emoji-cheat-sheet data from {url} has only {len(filtered)} entries"

    return filtered

def get_emoji_from_youtube(url: str) -> dict:
    """Get emoji alias from Youtube
    Returns a dict of emoji to list of short-names:
    E.g. {'💁': [':person_tipping_hand:', ':information_desk_person:'], '😉': [':winking_face:', ':wink:']}
    """

    data = requests.get(url).json()

    output = {}
    for obj in data:
        if 'shortcuts' not in obj or 'emojiId' not in obj:
            continue

        shortcuts = [x for x in obj['shortcuts'] if x.startswith(':') and x.endswith(':')]

        if shortcuts:
            output[obj['emojiId']] = shortcuts

    assert len(output) > 100, f"youtube data from {url} has only {len(output)} entries"

    return output



def extract_emojis(emojis_lines: list, sequences_lines: list) -> dict:
    """Extract emojis line by line to dict"""

    output = {}
    for line in emojis_lines:
        if not line == "" and not line.startswith("#"):
            emoji_status = line.split(";")[1].strip().split(" ")[0]

            codes = line.split(";")[0].strip().split(" ")
            separated_line = line.split(" # ")[-1].strip().split(" ")
            separated_name = separated_line[2:]
            version_str = separated_line[1]
            emoji_name = (
                "_".join(separated_name)
                .removeprefix("flag:_")
                .replace(":", "")
                .replace(",", "")
                .replace("\u201c", "")
                .replace("\u201d", "")
                .replace("\u229b", "")
                .strip()
                .replace(" ", "_")
                .replace("_-_", "-")
            )

            emoji_code = "".join(
                [
                    "\\U0000" + code if len(code) == 4 else "\\U000" + code
                    for code in codes
                ]
            )

            version = float(version_str.replace("E", "").strip())

            if emoji_code in output:
                raise Exception("Duplicate emoji: " +
                                emoji_name + " " + emoji_code)

            output[emoji_code] = {
                "en": emoji_name,
                "status": emoji_status.replace("-", "_"),
                "version": version
            }

    # Walk through the emoji-variation-sequences.txt
    for line in sequences_lines:
        if not line == "" and not line.startswith("#"):
            # No variant
            normal_codes = line.split(";")[0].strip().split(" ")
            normal_code = "".join(
                [
                    "\\U0000" + code if len(code) == 4 else "\\U000" + code
                    for code in normal_codes
                ]
            )
            if normal_code in output:
                output[normal_code]["variant"] = True

            # Text variant U+FE0E
            text_codes = re.sub(r'\s*FE0E\s*$', '',
                                line.split(";")[0]).strip().split(" ")
            text_code = "".join(
                [
                    "\\U0000" + code if len(code) == 4 else "\\U000" + code
                    for code in text_codes
                ]
            )
            if text_code in output:
                output[text_code]["variant"] = True

            # Emoji variant U+FE0F
            emoji_codes = re.sub(r'\s*FE0F\s*$', '', line.split(";")[0]).strip().split(" ")
            emoji_code = "".join(
                [
                    "\\U0000" + code if len(code) == 4 else "\\U000" + code
                    for code in emoji_codes
                ]
            )
            if emoji_code in output:
                output[emoji_code]["variant"] = True

    return output


def get_UNICODE_EMOJI(lang):
    return {emj: emoji_pkg.EMOJI_DATA[emj][lang] for emj in emoji_pkg.EMOJI_DATA if lang in emoji_pkg.EMOJI_DATA[emj]}


def adapt_emoji_name(text: str, lang: str, emj: str) -> str:
    # Use NFKC-form (single character instead of character + diacritic)
    # Unicode.org files should be formatted like this anyway, but emojiterra is not consistent
    text = unicodedata.normalize('NFKC', text)

    # Fix German clock times "12:30 Uhr" -> "12.30 Uhr"
    text = re.sub(r"(\d+):(\d+)", r"\1.\2", text)

    # Remove white space
    text = "_".join(text.split(" "))

    emoji_name = ":" + (
        text
        .lower()
        .removeprefix("flag:_")
        .replace(":", "")
        .replace(",", "")
        .replace('"', "")
        .replace("\u201e", "")
        .replace("\u201f", "")
        .replace("\u202f", "")
        .replace("\u229b", "")
        .replace("\u2013", "-")
        .replace(",_", ",")
        .strip()
        .replace(" ", "_")
        .replace("_-_", "-")
    ) + ":"

    if lang == "de":
        emoji_name = emoji_name.replace("\u201c", "").replace("\u201d", "")
        emoji_name = re.sub(r"(hautfarbe)_und_([a-z]+_hautfarbe)", r"\1,\2", emoji_name)

    if lang == "fa":
        emoji_name = emoji_name.replace('\u200c', "_")
        emoji_name = emoji_name.replace('\u200f', "_")
        emoji_name = emoji_name.replace('\u060c', "_")
        emoji_name = re.sub("_+", "_", emoji_name)

    if lang == "zh":
        emoji_name = ":" + (
            text
            .replace(":", "")
            .replace(",", "")
            .replace('-', "")
            .replace("\u201e", "")
            .replace("\u201f", "")
            .replace("\u202f", "")
            .replace("\u229b", "")
            .replace(",_", ",")
            .strip()
            .replace(" ", "_")
        ) + ":"

        if '日文' in emoji_name:
            # Japanese buttons
            emoji_name = emoji_name.replace('日文的', '').replace('按钮', '').replace('“', '').replace('”', '')

        if '箭头' in emoji_name:
            # Arrows
            emoji_name = emoji_name.replace('_', '').replace('!', '')

        if '按钮' in emoji_name:
            # English buttons
            emoji_name = emoji_name.replace('_', '')

        if '型血' in emoji_name:
            emoji_name = emoji_name.replace('_', '')

        if '中等-' in emoji_name:
            emoji_name = emoji_name.replace('中等-', '中等')

        if emoji_name.startswith(':旗_'):
            # Countries
            emoji_name = emoji_name.replace(':旗_', ':')

        hardcoded = {
            '\U0001f1ed\U0001f1f0': ':香港:',  # 🇭🇰
            '\U0001f1ee\U0001f1e9': ':印度尼西亞:',  # 🇮🇩
            '\U0001f1f0\U0001f1ff': ':哈薩克:',  # 🇰🇿
            '\U0001f1f2\U0001f1f4': ':澳門:',  # 🇲🇴
            '\U0001f1e8\U0001f1ec': ':刚果_布:',  # 🇨🇬
            '\U0001f1e8\U0001f1e9': ':刚果_金:',  # 🇨🇩
            '\U0001f193': ':FREE按钮:',  # 🆓
            '\U0001f238': ':申:',  # 🈸
            '\U0001f250': ':得:',  # 🉐
            '\U0001f22f': ':指:',  # 🈯
            '\U0001f232': ':禁:',  # 🈲
            '\u3297\ufe0f': ':祝:',  # ㊗️
            '\u3297': ':祝:',  # ㊗
            '\U0001f239': ':割:',  # 🈹
            '\U0001f21a': ':无:',  # 🈚
            '\U0001f237\ufe0f': ':月:',  # 🈷️
            '\U0001f237': ':月:',  # 🈷
            '\U0001f235': ':满:',  # 🈵
            '\U0001f236': ':有:',  # 🈶
            '\U0001f234': ':合:',  # 🈴
            '\u3299\ufe0f': ':秘:',  # ㊙️
            '\u3299': ':秘:',  # ㊙
            '\U0001f233': ':空:',  # 🈳
            '\U0001f251': ':可:',  # 🉑
            '\U0001F23A': ':营:',  # 🈺
            '\U0001F202\ufe0f': ':服务:',  # 🈂️
            '\U0001F202': ':服务:',  # 🈂
        }

        if emj in hardcoded:
            emoji_name = hardcoded[emj]

    emoji_name = (emoji_name
        .replace("____", "_")
        .replace("___", "_")
        .replace("__", "_")
        .replace("--", "-"))

    return emoji_name


def add_unicode_annotations(data, lang, url):
    xml = get_text_from_url(url)

    tree = ET.fromstring(xml)
    annotations = tree.find('annotations')
    for annotation in annotations:
        if annotation.get('type') == 'tts':
            emj = annotation.get('cp')
            text = annotation.text.strip()

            emoji_name = adapt_emoji_name(text, lang, emj)

            if emj in data and data[emj] != emoji_name:
                print(
                    f"# {lang}: CHANGED {data[emj]} TO {emoji_name} \t\t(Original: {text})")
            data[emj] = emoji_name


def extract_names(github_tag, github_lang, lang, emoji_terra={}):
    """Copies emoji.EMOJI_DATA[emj][lang] and adds the names from the Unicode CLDR xml

    Find latest tag at https://cldr.unicode.org/index/downloads or
    https://github.com/unicode-org/cldr/tree/main/common/annotations
    """

    data = get_UNICODE_EMOJI(lang)
    add_unicode_annotations(data, lang, f"https://github.com/unicode-org/cldr/raw/{github_tag}/common/annotations/{github_lang}.xml")
    add_unicode_annotations(data, lang, f"https://github.com/unicode-org/cldr/raw/{github_tag}/common/annotationsDerived/{github_lang}.xml")

    # Add names from emojiterra if there is no unicode annotation
    for emj, name in emoji_terra.items():
        if emj in emoji_pkg.EMOJI_DATA and emj not in data:
            emoji_name = adapt_emoji_name(name, lang, emj)
            data[emj] = emoji_name

    # There are some emoji with two code sequences for the same emoji, one that ends with \uFE0F and one that does not.
    # The one that ends with \uFE0F is the "new" emoji, that is RGI.
    # The Unicode translation data sometimes only has one of the two code sequences and is missing the other one.
    # In that case we want to use the existing translation for both code sequences.
    missing_translation = {}
    for emj in data:
        if emj.endswith('\uFE0F') and emj[0:-1] not in data and emj[0:-1] in emoji_pkg.EMOJI_DATA:
            # the emoji NOT ending in \uFE0F exists in EMOJI_DATA but is has no translation
            # e.g. ':pirate_flag:' -> '\U0001F3F4\u200D\u2620\uFE0F' or '\U0001F3F4\u200D\u2620'
            missing_translation[emj[0:-1]] = data[emj]

        with_emoji_type = f"{emj}\uFE0F"
        if not emj.endswith('\uFE0F') and with_emoji_type not in data and with_emoji_type in emoji_pkg.EMOJI_DATA:
            # the emoji ending in \uFE0F exists in EMOJI_DATA but is has no translation
            # e.g. ':face_in_clouds:' -> '\U0001F636\u200D\U0001F32B\uFE0F' or '\U0001F636\u200D\U0001F32B'
            missing_translation[with_emoji_type] = data[emj]

    # Find emoji that contain \uFE0F inside the sequence (not just as a suffix)
    # e.g. ':eye_in_speech_bubble:' -> '\U0001F441\uFE0F\u200D\U0001F5E8\uFE0F'
    for emj in emoji_pkg.EMOJI_DATA:
        if emj in data:
            continue
        emj_no_variant = emj.replace('\uFE0F', '')
        if emj_no_variant != emj and emj_no_variant in data:
            # the emoji with \uFE0F has no translation, but the emoji without all \uFE0F has a translation
            data[emj] = data[emj_no_variant]

    data.update(missing_translation)

    return data


def get_emoji_from_github_api(url: str) -> dict:
    """Get emoji alias from GitHub API
    """

    data = requests.get(url).json()
    pattern = re.compile(r"unicode/([0-9a-fA-F-]+)\.[a-z]+")

    output = {}
    for name, img in data.items():
        m = pattern.search(img)
        if m:
            emj = "".join(chr(int(h, 16)) for h in m.group(1).split('-'))
            output[name] = emj
        else:
            pass  # Special GitHub emoji that is not part of Unicode

    assert len(output) > 100, f"data from github API has only {len(output)} entries"

    return output


GITHUB_REMOVED_CHARS = re.compile("\u200D|\uFE0F|\uFE0E", re.IGNORECASE)


def find_github_aliases(emj, github_dict, v, emj_no_variant=None):
    aliases = set()

    # Strip ZWJ \u200D, text_type \uFE0E and emoji_type \uFE0F
    # because the GitHub API does not include these
    emj_clean = GITHUB_REMOVED_CHARS.sub("", emj)

    for gh_alias in github_dict:
        if emj == github_dict[gh_alias]:
            aliases.add(gh_alias)
        elif 'variant' in v and emj_no_variant == github_dict[gh_alias]:
            aliases.add(gh_alias)
        elif emj_clean == github_dict[gh_alias]:
            aliases.add(gh_alias)

    return aliases


def ascii(s):
    # return escaped Code points \U000AB123
    return s.encode("unicode-escape").decode()


if __name__ == "__main__":
    logging.info('  Downloading...\n')

    # Find the latest version at https://www.unicode.org/reports/tr51/#emoji_data
    emoji_source = get_emoji_from_url(15.0)
    emoji_sequences_source = get_emoji_variation_sequence_from_url('15.0.0')
    emojis = extract_emojis(emoji_source, emoji_sequences_source)
    # Find latest release tag at https://cldr.unicode.org/index/downloads
    github_tag = 'release-43-1'

    languages = {
        # Update names in other languages:
        'de': extract_names(github_tag, 'de', 'de', get_emojiterra_from_url('https://emojiterra.com/de/kopieren/')),
        'es': extract_names(github_tag, 'es', 'es', get_emojiterra_from_url('https://emojiterra.com/es/copiar/')),
        'fr': extract_names(github_tag, 'fr', 'fr', get_emojiterra_from_url('https://emojiterra.com/fr/copier/')),
        'ja': extract_names(github_tag, 'ja', 'ja', get_emojiterra_from_url('https://emojiterra.com/copypaste/ja/')),
        'ko': extract_names(github_tag, 'ko', 'ko', get_emojiterra_from_url('https://emojiterra.com/copypaste/ko/')),
        'pt': extract_names(github_tag, 'pt', 'pt', get_emojiterra_from_url('https://emojiterra.com/pt/copiar/')),
        'it': extract_names(github_tag, 'it', 'it', get_emojiterra_from_url('https://emojiterra.com/it/copiare/')),
        'fa': extract_names(github_tag, 'fa', 'fa', get_emojiterra_from_url('https://emojiterra.com/copypaste/fa/')),
        'id': extract_names(github_tag, 'id', 'id', get_emojiterra_from_url('https://emojiterra.com/copypaste/id/')),
        'zh': extract_names(github_tag, 'zh', 'zh', get_emojiterra_from_url('https://emojiterra.com/copypaste/zh/')),

        # Do not update names in other languages:
        # 'de': get_UNICODE_EMOJI('de'),
        # 'es': get_UNICODE_EMOJI('es'),
        # 'fr': get_UNICODE_EMOJI('fr'),
        # 'ja': get_UNICODE_EMOJI('ja'),
        # 'ko': get_UNICODE_EMOJI('ko'),
        # 'pt': get_UNICODE_EMOJI('pt'),
        # 'it': get_UNICODE_EMOJI('it'),
        # 'fa': get_UNICODE_EMOJI('fa'),
        # 'id': get_UNICODE_EMOJI('id'),
        # 'zh': get_UNICODE_EMOJI('zh'),
    }

    github_alias_dict = get_emoji_from_github_api('https://api.github.com/emojis')
    cheat_sheet_dict = get_cheat_sheet('https://www.webfx.com/tools/emoji-cheat-sheet/')
    youtube_dict = get_emoji_from_youtube('https://www.gstatic.com/youtube/img/emojis/emojis-png-7.json')

    logging.info('  Combining...\n')

    used_github_aliases = set()

    escapedToUnicodeMap = {escaped: escaped.encode().decode('unicode-escape') for escaped in emojis}  # maps: "\\U0001F4A4" to "\U0001F4A4"

    all_existing_aliases_and_en = set(item for emj_data in emoji_pkg.EMOJI_DATA.values() for item in emj_data.get('alias', []))
    all_existing_aliases_and_en.update(emj_data['en'] for emj_data in emoji_pkg.EMOJI_DATA.values())

    f = 0
    c = 0
    new_aliases = []
    logging.info('  Print EMOJI_DATA...\n')
    for code, v in sorted(emojis.items(), key=lambda item: item[1]["en"]):
        language_str = ''
        emj = escapedToUnicodeMap[code]

        alternative = re.sub(r"\\U0000FE0[EF]$", "", code)
        emj_no_variant = escapedToUnicodeMap[alternative]

        # add names in other languages
        for lang in languages:
            if emj in languages[lang]:
                language_str += ",\n        '%s': '%s'" % (
                    lang, languages[lang][emj])
            elif 'variant' in v:
                # the language annotation uses the normal emoji (no variant), while the emoji-test.txt uses the emoji or text variant
                if emj_no_variant in languages[lang]:
                    language_str += ",\n        '%s': '%s'" % (
                        lang, languages[lang][emj_no_variant])

        # Add existing alias from EMOJI_DATA
        aliases = set()
        if emj in emoji_pkg.EMOJI_DATA and 'alias' in emoji_pkg.EMOJI_DATA[emj]:
            aliases.update(a[1:-1] for a in emoji_pkg.EMOJI_DATA[emj]['alias'])
        old_aliases = set(aliases)

        if emj_no_variant in emoji_pkg.EMOJI_DATA and 'alias' in emoji_pkg.EMOJI_DATA[emj_no_variant]:
            aliases.update(a[1:-1] for a in emoji_pkg.EMOJI_DATA[emj_no_variant]['alias'])

        # Add alias from GitHub API
        github_aliases = find_github_aliases(emj, github_alias_dict, v, emj_no_variant)
        aliases.update(shortcut for shortcut in github_aliases if shortcut not in all_existing_aliases_and_en)
        used_github_aliases.update(github_aliases)

        # Add alias from cheat sheet
        if emj in cheat_sheet_dict and cheat_sheet_dict[emj] not in all_existing_aliases_and_en:
            aliases.add(cheat_sheet_dict[emj][1:-1])
        if emj_no_variant in cheat_sheet_dict and cheat_sheet_dict[emj_no_variant] not in all_existing_aliases_and_en:
            aliases.add(cheat_sheet_dict[emj_no_variant][1:-1])

        # Add alias from youtube
        if emj in youtube_dict:
            aliases.update(shortcut[1:-1] for shortcut in youtube_dict[emj] if shortcut not in all_existing_aliases_and_en)
        if emj_no_variant in youtube_dict:
            aliases.update(shortcut[1:-1] for shortcut in youtube_dict[emj_no_variant] if shortcut not in all_existing_aliases_and_en)

        # Remove if alias is same as 'en'-name
        if v["en"] in aliases:
            aliases.remove(v["en"])

        # Store new aliases to print them at the end after the dict of dicts
        if emj in emoji_pkg.EMOJI_DATA:
            if 'alias' in emoji_pkg.EMOJI_DATA[emj]:
                diff = aliases.difference(a[1:-1] for a in emoji_pkg.EMOJI_DATA[emj]['alias'])
            else:
                diff = aliases
            for a in diff:
                new_aliases.append(f"# alias NEW {a} FOR {emj} CODE {code}")

        # Keep the order of existing aliases intact
        if emj in emoji_pkg.EMOJI_DATA and 'alias' in emoji_pkg.EMOJI_DATA[emj]:
            aliases = [a[1:-1] for a in emoji_pkg.EMOJI_DATA[emj]['alias']] + [a for a in aliases if f":{a}:" not in emoji_pkg.EMOJI_DATA[emj]['alias']]

        if any("flag_for_" in a for a in aliases):
            # Put the :flag_for_COUNTRY: alias as the first entry so that it gets picked by demojize()
            # This ensures compatibility because in the past there was only the :flag_for_COUNTRY: alias
            aliases = [a for a in aliases if "flag_for_" in a] + [a for a in aliases if "flag_for_" not in a]

        # Print dict of dicts
        alias = ''
        if len(aliases) > 0:
            alias_list_str = ", ".join([f"':{a}:'" for a in aliases])
            alias = ",\n        'alias': [%s]" % (alias_list_str, )
        variant = ",\n        'variant': True" if 'variant' in v else ''
        print(f"""    '{code}': {{  # {emj}
        'en': ':{v['en']}:',
        'status': {v["status"]},
        'E': {v["version"]:g}{alias}{variant}{language_str}
    }}""", end=",\n")
        if v["status"] == "fully_qualified":
            f += 1
        elif v["status"] == "component":
            c += 1

    logging.debug(f" # Total count of emojis: {len(emojis)}")
    logging.debug(f" # fully_qualified: {f}")
    logging.debug(f" # component: {c}\n")
    logging.debug("\n".join(new_aliases))

    # Check if all aliases from GitHub API were used
    for github_alias in github_alias_dict:
        if github_alias not in used_github_aliases:
            logging.debug(f"# Unused Github alias: {github_alias} {github_alias_dict[github_alias]} {ascii(github_alias_dict[github_alias])}")

    logging.info('\n\n  Done.')


"""Unittests for emoji.analyze()"""


import emoji


def test_analyze():
    assert list(emoji.analyze('')) == []

    assert list(emoji.analyze('', non_emoji=True)) == []

    assert list(emoji.analyze('abc')) == []

    assert list(emoji.analyze('abc', non_emoji=True)) == [('a', 'a'), ('b', 'b'), ('c', 'c')]

    result = list(emoji.analyze('abc\U0001F472'))
    assert len(result) == 1
    assert result[0].value.emoji == '\U0001F472'

    result = list(emoji.analyze('abc\U0001F472', non_emoji=True))
    assert result[0].value == 'a'
    assert result[3].value.emoji == '\U0001F472'

    result = list(emoji.analyze('\U0001F477\U0001F3FB\U0000200D\U00002640'))
    assert len(result) == 1
    assert result[0].value.emoji == '\U0001F477\U0001F3FB\U0000200D\U00002640'

    result = list(emoji.analyze('\U0001F477\U0001F3FC\U0001F477\U0001F3FB\U0000200D\U00002640'))
    assert len(result) == 2
    assert result[0].value.emoji == '\U0001F477\U0001F3FC'
    assert result[1].value.emoji == '\U0001F477\U0001F3FB\U0000200D\U00002640'


def test_analyze_non_rgi_zwj():
    result = list(emoji.analyze('\U0001F468\U0001F3FF\U0000200D\U0001F469\U0001F3FB\U0000200D\U0001F467\U0001F3FD'))
    assert len(result) == 1
    assert result[0].value.emoji == '\U0001F468\U0001F3FF\U0000200D\U0001F469\U0001F3FB\U0000200D\U0001F467\U0001F3FD'

    result = list(emoji.analyze('\U0001F468\U0001F3FF\U0000200D\U0001F469\U0001F3FB\U0000200D\U0001F467\U0001F3FD', join_emoji=False))
    assert len(result) == 3
    assert result[0].value.emoji == '\U0001F468\U0001F3FF'
    assert result[1].value.emoji == '\U0001F469\U0001F3FB'
    assert result[2].value.emoji == '\U0001F467\U0001F3FD'

    result = list(emoji.analyze('\U0001F468\U0001F3FF\U0000200D\U0001F469\U0001F3FB\U0000200D\U0001F467\U0001F3FDx', join_emoji=False, non_emoji=True))
    assert len(result) == 6
    assert result[0].value.emoji == '\U0001F468\U0001F3FF'
    assert result[1].value == '\U0000200D'
    assert result[2].value.emoji == '\U0001F469\U0001F3FB'
    assert result[3].value == '\U0000200D'
    assert result[4].value.emoji == '\U0001F467\U0001F3FD'
    assert result[5].value == 'x'

    result = list(emoji.analyze('\U0001F468\U0001F3FF\U0000200D\U0001F469\U0001F3FB\U0000200D\U0001F467\U0001F3FDx', join_emoji=True, non_emoji=True))
    assert len(result) == 2
    assert result[0].value.emoji == '\U0001F468\U0001F3FF\U0000200D\U0001F469\U0001F3FB\U0000200D\U0001F467\U0001F3FD'
    assert result[1].value == 'x'

    result = list(emoji.analyze("\u200D🦷\u200D🦷"))
    assert len(result) == 1
    assert isinstance(result[0].value, emoji.EmojiMatchZWJNonRGI)

    result = list(emoji.analyze("\u200D🦷\u200D🦷", join_emoji=False))
    assert len(result) == 2
    assert all(isinstance(token.value, emoji.EmojiMatch) for token in result)

    result = list(emoji.analyze("\u200D🦷\u200D🦷", join_emoji=False, non_emoji=True))
    assert len(result) == 4
    assert result[0].value == '\u200D'
    assert isinstance(result[1].value, emoji.EmojiMatch)
    assert result[2].value == '\u200D'
    assert isinstance(result[3].value, emoji.EmojiMatch)


def test_emoji_match():
    s = 'a\U0001F309b'
    token = next(emoji.analyze(s))
    assert isinstance(token, emoji.Token)

    assert token.chars == s[1:-1]

    match = token.value

    assert isinstance(match, emoji.EmojiMatch)
    assert match.emoji == s[1:-1]
    assert match.start == 1
    assert match.end == 2
    assert match.is_zwj() == False
    assert str(match).startswith('EmojiMatch(')


def test_emoji_match():
    s = 'a\U0001F309b'
    token = next(emoji.analyze(s))
    assert isinstance(token, emoji.Token)

    assert token.chars == s[1:-1]

    match = token.value

    assert isinstance(match, emoji.EmojiMatch)
    assert match.emoji == s[1:-1]
    assert match.start == 1
    assert match.end == 2
    assert match.is_zwj() == False
    assert match.split() == match
    assert str(match).startswith('EmojiMatch(')


    s = 'a\U0001F468\U0000200D\U0001F467b'
    token = next(emoji.analyze(s))
    assert isinstance(token, emoji.Token)

    assert token.chars == s[1:-1]

    match = token.value

    assert isinstance(match, emoji.EmojiMatch)
    assert match.emoji == s[1:-1]
    assert match.start == 1
    assert match.end == 4
    assert match.is_zwj() == True
    assert str(match).startswith('EmojiMatch(')

    splitted = match.split()
    assert isinstance(splitted, emoji.EmojiMatchZWJ)
    assert splitted.emoji == s[1:-1]
    assert splitted.start == 1
    assert splitted.end == 4
    assert splitted.is_zwj() == True
    assert splitted.join() == s[1:-1]
    assert splitted.split() == splitted
    assert str(splitted).startswith('EmojiMatchZWJ(')
    assert len(splitted.emojis) == 2

    man, woman = splitted.emojis

    assert isinstance(man, emoji.EmojiMatch)
    assert man.start == 1
    assert man.end == 2

    assert isinstance(woman, emoji.EmojiMatch)
    assert woman.start == 3
    assert woman.end == 4



"""Generate the files index.html and all.html that will contain a table of all supported emoji
Run with -minify to minify HTML for production"""

import sys
import os
import codecs
import django.conf
import django.template
from htmlmin.minify import html_minify

try:
    import emoji
except ImportError:
    include = os.path.relpath(os.path.join(os.path.dirname(__file__), "../.."))
    sys.path.insert(0, include)
    import emoji
    print("Imported emoji from %s" %
          os.path.abspath(os.path.join(include, "emoji")))

# Configuration
OUT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMPLATE_DIR = os.path.abspath(os.path.dirname(__file__))
TEMPLATE_FILE = os.path.join(TEMPLATE_DIR, "template.html")
data = {"defaultLang": "en"}

languages = {
    "en": emoji.emojize("en :United_Kingdom:"),
    "alias":  emoji.emojize("alias :United_Kingdom:"),
    "es": emoji.emojize("es :Spain:"),
    "ja": emoji.emojize("ja :Japan:"),
    "ko": emoji.emojize("ko :South_Korea:"),
    "pt": emoji.emojize("pt :Portugal:"),
    "it": emoji.emojize("it :Italy:"),
    "fr": emoji.emojize("fr :France:"),
    "de": emoji.emojize("de :Germany:"),
    "fa": emoji.emojize("fa :Iran:"),
    "id": emoji.emojize("id :Indonesia:"),
    "zh": emoji.emojize("zh :China:"),
}
language_args = {}


minify_enabled = sys.argv[-1] == '-minify'
if not minify_enabled:
    print("Run with -minify to minify HTML")


print("Collecting emoji data...")

data["lists"] = lists = []
for language in languages:

    if language in language_args:
        language_arg = language_args[language]
    elif language == data["defaultLang"]:
        language_arg = ''
    else:
        language_arg = f'language="{language}"'

    emoji_list = []
    for code, emoji_data in emoji.EMOJI_DATA.items():
        if language not in emoji_data:
            continue
        names = [emoji_data[language]] if isinstance(emoji_data[language], str) else emoji_data[language]
        for name in names:
            emoji_list.append({
                "code": code,
                "name": name,
                "unicode": code.encode('ascii', 'backslashreplace').decode('ascii'),
                "charname": code.encode('ascii', 'namereplace').decode('ascii'),
                "xml": code.encode('ascii', 'xmlcharrefreplace').decode('ascii')
            })

    listentry = {
        "name": language,
        "pretty": languages[language],
        "languageArg": language_arg,
        "emojis": emoji_list
    }
    lists.append(listentry)


# Render with django
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [TEMPLATE_DIR],
    }
]

django.conf.settings.configure(TEMPLATES=TEMPLATES)
django.setup()
template = django.template.loader.get_template(TEMPLATE_FILE)

print("Render 'all.html' ...")
with codecs.open(os.path.join(OUT_DIR, "all.html"), "w", encoding="utf-8") as f:
    html = template.render(data)
    if minify_enabled:
        print("Minify 'all.html' ...")
        html = html_minify(html)
    f.write(html)
    print(f"Wrote to {os.path.join(OUT_DIR, 'all.html')}")

print("Render 'index.html' ...")
# Remove emoji except for default list
for listentry in lists:
    if listentry["name"] != data["defaultLang"]:
        listentry["emojis"] = []

with codecs.open(os.path.join(OUT_DIR, "index.html"), "w", encoding="utf-8") as f:
    html = template.render(data)
    if minify_enabled:
        print("Minify 'index.html' ...")
        html = html_minify(html)
    f.write(html)
    print(f"Wrote to {os.path.join(OUT_DIR, 'index.html')}")

print("Done.")



#!/usr/bin/env python


"""Setup script for emoji."""


import os
from codecs import open

from setuptools import setup

with open('README.rst', encoding='utf-8') as f:
    readme_content = f.read().strip()


author = email = source = version = None
with open(os.path.join('emoji', '__init__.py'), encoding='utf-8') as f:
    for line in f:
        if line.strip().startswith('__version__'):
            version = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__author__'):
            author = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__email__'):
            email = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__source__'):
            source = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif None not in (version, author, email, source):
            break

setup(
    name='emoji',
    author=author,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    author_email=email,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Multimedia :: Graphics :: Presentation',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Typing :: Typed'
    ],
    description='Emoji for Python',
    keywords=['emoji'],
    extras_require={
        'dev': [
            'pytest',
            'coverage',
            'coveralls',
        ],
    },
    include_package_data=True,
    license='New BSD',
    long_description=readme_content,
    packages=['emoji', 'emoji.unicode_codes'],
    package_data={
        'emoji': [
            'py.typed',
            '*.pyi',
        ],
        'emoji.unicode_codes': [
            'py.typed',
            '*.pyi',
        ],
    },
    url=source,
    version=version,
    zip_safe=True,
)
